#!/usr/bin/env python
"""\
A service that just keeps bzrlib loaded in memory.
"""

import bzrlib, bzrlib.commands, bzrlib.builtins
import socket
import struct

from SocketServer import ForkingTCPServer, StreamRequestHandler

class BzrRequestHandler(StreamRequestHandler):
    def handle(self):
        import os
        # the first line is the directory to run in
        bytes = struct.unpack('>I', self.rfile.read(4))[0]

        active_dir = self.rfile.read(bytes)
        os.chdir(active_dir)
        # The command to run is newline separated
        nargs = struct.unpack('>I', self.rfile.read(4))[0]

        args = []
        for i in xrange(nargs):
            bytes = struct.unpack('>I', self.rfile.read(4))[0]
            args.append(self.rfile.read(bytes))

        bzrlib.trace.log_startup(args)
        retcode, stdout = run_bzr_captured(args)

        self.wfile.write(struct.pack('>I', retcode))
        self.wfile.flush()

        self.wfile.write(struct.pack('>i', len(stdout)))
        self.wfile.write(stdout)
        self.wfile.flush()

class BzrTCPServer(ForkingTCPServer):
    allow_reuse_address = True

def run_bzr_captured(argv):
    """Invoke bzr and return (result, stdout_and_err).

    Useful for code that wants to check the contents of the
    output, the way error messages are presented, etc.

    argv -- arguments to invoke bzr
    retcode -- expected return code, or None for don't-care.
    """
    from StringIO import StringIO
    import logging

    stdout_and_err = StringIO()
    #handler = logging.StreamHandler(stdout_and_err)
    #handler.setFormatter(bzrlib.trace.QuietFormatter())
    #handler.setLevel(logging.INFO)
    #logger = logging.getLogger('')
    #logger.addHandler(handler)
    try:
        result = apply_redirected(None, stdout_and_err,
                                  bzrlib.commands.run_bzr_catch_errors,
                                  argv)
    finally:
        #logger.removeHandler(handler)
        pass
    return result, stdout_and_err.getvalue()

def apply_redirected(stdin=None, stdout_and_err=None,
                     a_callable=None, *args, **kwargs):
    """Call callable with redirected std io pipes.

    Returns the return code.
    """
    from StringIO import StringIO
    import sys

    if not callable(a_callable):
        raise ValueError("a_callable must be callable.")
    if stdin is None:
        stdin = StringIO("")
    real_stdin = sys.stdin
    real_stdout = sys.stdout
    real_stderr = sys.stderr
    try:
        sys.stdout = stdout_and_err
        sys.stderr = stdout_and_err
        sys.stdin = stdin
        return a_callable(*args, **kwargs)
    finally:
        sys.stdout = real_stdout
        sys.stderr = real_stderr
        sys.stdin = real_stdin

class cmd_service(bzrlib.commands.Command):
    """Run a bzr service on the local machine, which should make startup time faster."""
    def run(self, *args, **kwargs):
        # Try to load all of the code right now,
        # so that the spawned program won't have to
        try:
            import bzrlib.transport.local
            import bzrlib.transport.http
        except ImportError:
            pass # Not all versions have Transports

        import bzrlib.branch
        import bzrlib.weave, bzrlib.weavefile
        import bzrlib.inventory
        import bzrlib.trace

        import bzrlib.ui
        import bzrlib.xml
        try:
            import bzrlib.xml4
            import bzrlib.xml5
        except ImportError:
            pass

        import os
        import sys
        import re
        import time
        import cStringIO
        import StringIO
        import logging

        bzrlib.ui.ui_factory = bzrlib.ui.TextUIFactory()

        server = BzrTCPServer(('localhost', 11111), BzrRequestHandler)

        server.serve_forever()

bzrlib.commands.register_command(cmd_service)

